(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import sepalg.
Require Import sepalg_generators.
Require Import boolean_alg.
Require Import cross_split.
Require tree_shares.

Module Share : SHARE_MODEL := tree_shares.Share.
Import Share.

Theorem leq_join_sub : forall s1 s2:Share.t,
  s1 <= s2 <-> join_sub s1 s2.
Proof.
  split; intros.
  pose (s' := glb s2 (comp s1)).
  exists s'.
  simpl; split.
  subst s'.
  rewrite glb_commute.
  rewrite glb_assoc.
  rewrite (glb_commute (comp s1) s1).
  rewrite comp2.
  apply glb_bot.
  subst s'.
  rewrite distrib2.
  rewrite comp1.
  rewrite glb_top.
  rewrite <- ord_spec2; auto.

  destruct H as [s' H].
  destruct H.
  rewrite ord_spec2.
  rewrite <- H0.
  rewrite <- lub_assoc.
  rewrite lub_idem; auto.
Qed.

Lemma top_correct' : forall x:t, join_sub x top.
Proof.
  intros; rewrite <- leq_join_sub; auto with ba.
Qed.

Lemma bot_identity : identity bot.
Proof.
  hnf; intros.
  destruct H.
  rewrite lub_commute in H0.
  rewrite lub_bot in H0.
  auto.
Qed.

Lemma identity_share_bot : forall s,
  identity s -> s = bot.
Proof.
  intros.
  apply identities_unique; auto.
  apply bot_identity.
  exists s.
  apply join_com.
  
  destruct (top_correct' s).
  assert (x = top).
  apply H; auto.
  subst x; auto.
  destruct (top_correct' bot).
  assert (x = top).
  apply bot_identity; auto.
  subst x; auto.
  apply join_com in H1.
  destruct (join_assoc H0 H1); intuition.
  assert (x = top).
  apply H; auto.
  subst x.
  replace bot with s.
  rewrite identity_unit_equiv in H.
  trivial.
  apply joins_units_eq with top bot; firstorder.
Qed.

Lemma factoryOverlap' : forall f1 f2 n1 n2,
  isTokenFactory f1 n1 -> isTokenFactory f2 n2 -> joins f1 f2 -> False.
Proof.
  intros.
  destruct H1.
  destruct H1.
  apply (factoryOverlap f1 f2 n1 n2 H H0 H1).
Qed.

Lemma identityToken' : forall x, isToken x 0 <-> identity x.
Proof.
  intro x; destruct (identityToken x); split; intros.
  hnf; intros.
  rewrite H in H2; auto.
  apply bot_identity; auto.
  apply H0.
  apply identity_share_bot; auto.
Qed.

Lemma nonidentityToken' : forall x n, (n > 0)%nat -> isToken x n -> nonidentity x.
Proof.
  intros.
  generalize (nonidentityToken x n H H0).
  repeat intro.
  apply H1.
  apply identity_share_bot; auto.
Qed.

Lemma nonidentityFactory' : forall x n, isTokenFactory x n -> nonidentity x.
Proof.
  intros.
  generalize (nonidentityFactory x n H); repeat intro.
  apply H0.
  apply identity_share_bot; auto.
Qed.

Lemma split_join : forall x1 x2 x,
  split x = (x1,x2) -> join x1 x2 x.
Proof.
  intros; split.
  apply split_disjoint with x; auto.
  apply split_together; auto.
Qed.

Lemma split_nontrivial' : forall x1 x2 x,
  split x = (x1, x2) ->
    (identity x1 \/ identity x2) ->
    identity x.
Proof.
  intros.
  rewrite (split_nontrivial x1 x2 x H).
  apply bot_identity.
  destruct H0.
  left; apply identity_share_bot; auto.
  right; apply identity_share_bot; auto.
Qed.

Lemma rel_leq : forall a x, join_sub (rel a x) a.
Proof.
  intros.
  rewrite <- leq_join_sub.

  intros.
  rewrite ord_spec1.
  pattern a at 3.
  replace a with (rel a top).
  rewrite <- rel_preserves_glb.
  rewrite glb_top.
  auto.
  apply rel_top1.
Qed.

Lemma rel_join : forall a x y z,
  join x y z ->
  join (rel a x) (rel a y) (rel a z).
Proof.
  simpl; unfold ba_join; intuition.
  rewrite <- rel_preserves_glb.
  replace bot with (rel a bot).
  replace (glb x y) with bot; auto.
  apply rel_bot1.
  rewrite <- rel_preserves_lub.
  replace (lub x y) with z; auto.
Qed.

Lemma rel_join2 : forall a x y s,
  nonidentity a ->
  join (rel a x) (rel a y) s ->
  exists z, s = rel a z /\ join x y z.
Proof.
  simpl; intros.
  destruct H0.
  exists (lub x y).
  split.
  rewrite <- H1.
  symmetry.
  apply rel_preserves_lub.
  split; auto.
  rewrite <- rel_preserves_glb in H0.
  replace bot with (rel a bot) in H0.
  apply rel_inj_l with a; auto.
  hnf; intros; apply H.
  subst a; apply bot_identity.
  apply rel_bot1.
Qed.

Lemma rel_nontrivial : forall a x,
  identity (rel a x) ->
  (identity a \/ identity x).
Proof.
  intros a x H.
  destruct (eq_dec a bot); auto.
  subst a.
  left. apply bot_identity.

  right.
  assert (rel a x = bot).
  apply identity_share_bot; auto.
  assert (x = bot).
  replace bot with (rel a bot) in H0.
  apply rel_inj_l with a; auto.
  apply rel_bot1.
  subst x; apply bot_identity.
Qed.

Lemma share_cross_split : @cross_split t sa.
Proof.
  hnf; simpl; unfold ba_join; intuition.
  exists (glb a c, glb a d, glb b c, glb b d); intuition.
  rewrite (glb_commute a d).
  rewrite glb_assoc.
  rewrite <- (glb_assoc c d a).
  rewrite H.
  rewrite (glb_commute bot a).
  rewrite glb_bot.
  rewrite glb_bot.
  auto.
  rewrite <- distrib1;  rewrite H3; rewrite <- H2; auto with ba.
  rewrite (glb_commute b d).
  rewrite glb_assoc.
  rewrite <- (glb_assoc c d b).
  rewrite H.
  rewrite (glb_commute bot b).
  rewrite glb_bot.
  rewrite glb_bot.
  auto.
  rewrite <- distrib1;  rewrite H3; rewrite <- H2; auto with ba.
  rewrite (glb_commute a c).
  rewrite glb_assoc.
  rewrite <- (glb_assoc a b c).
  rewrite H1.
  rewrite (glb_commute bot c).
  rewrite glb_bot.
  rewrite glb_bot.
  auto.
  rewrite (glb_commute a c).
  rewrite (glb_commute b c).
  rewrite <- distrib1; rewrite H2; rewrite <- H3; auto with ba.
  rewrite (glb_commute a d).
  rewrite glb_assoc.
  rewrite <- (glb_assoc a b d).
  rewrite H1.
  rewrite (glb_commute bot d).
  rewrite glb_bot.
  rewrite glb_bot.
  auto.
  rewrite (glb_commute a d).
  rewrite (glb_commute b d).
  rewrite <- distrib1; rewrite H2; rewrite <- H3; auto with ba.
Qed.

Lemma bot_correct' : forall x, join_sub bot x.
Proof.
  intros s.
  destruct (top_correct' s).
  exists s.
  destruct (top_correct' bot).
  assert (x0 = top).
  apply bot_identity; auto.
  subst x0.
  apply join_com in H0.
  destruct (join_assoc H H0); intuition.
  apply join_com in H2.
  destruct (join_assoc H1 H2); intuition.
  assert (s = x1).
  apply bot_identity; auto.
  subst x1; auto.
Qed.

Lemma top_share_nonidentity : nonidentity top.
Proof.
  hnf; intros.
  assert (top = bot).
  apply identity_share_bot; auto.
  apply nontrivial; auto.
Qed.

Lemma bot_joins : forall x, joins bot x.
Proof.
  intro x; exists x.
  destruct (join_ex_identities x); intuition.
  destruct H0.
  generalize (H _ _ H0).
  intros; subst; auto.
  replace bot with x0; auto.
  apply identity_share_bot; auto.
Qed.

Lemma dec_share_identity : forall x:t, { identity x } + { ~identity x }.
Proof.
  intro x.
  destruct (Share.eq_dec x bot); subst.
  left; apply bot_identity.
  right; intro; elim n.
  apply identity_share_bot; auto.
Qed.

  (*  A pshare is a "positive share", that is, a nonidentity share. *)
Definition pshare := pjoinable sa.
Definition pshare_sh (psh: pshare) : t := proj1_sig psh.
Definition pshare_nonidentity (psh: pshare) : nonidentity (pshare_sh psh) := proj2_sig psh.

Coercion pshare_sh : pshare >-> t.
Definition mk_pshare (sh: t) (p: nonidentity sh) := exist nonidentity sh p.
Definition pfullshare : pshare := mk_pshare top top_share_nonidentity.

Lemma pshare_eq_e: forall (sh sh' : pshare), pshare_sh sh = pshare_sh sh' -> sh=sh'.
Proof.
  destruct sh; destruct sh'.
  simpl.
  intros; subst x0.
  rewrite (proof_irr n n0); auto.
Qed.
Implicit Arguments pshare_eq_e [sh sh'].

Theorem pshare_leq_join : forall s1 s2:pshare,
  join_sub (proj1_sig s1) (proj1_sig s2) -> joins (proj1_sig s1) (proj1_sig s2) -> False.
Proof.
  simpl; intros.
  generalize (join_sub_joins H H0); intros.
  destruct H1.
  assert (pshare_sh s1 = x).
  apply join_self; auto.
  subst x. 
  contradiction (pshare_nonidentity s1).
  destruct s1.
  apply unit_identity with x; auto.
Qed.

Lemma pshare_join_full_false1 : forall (p:pshare),
  joins (proj1_sig pfullshare) (proj1_sig p) -> False.
Proof.
  intros.
  contradiction (pshare_nonidentity p).
  rewrite joins_sym in H.
  generalize (join_sub_joins (top_correct' _) H); intros.
  destruct H0.
  assert (pshare_sh p = x).
  apply join_self; auto.
  subst x. 
  destruct p.
  apply unit_identity with x; auto.
Qed.

Lemma pshare_join_full_false2 : forall (p:pshare),
  joins (proj1_sig p) (proj1_sig pfullshare)  -> False.
Proof.
  intros.
  destruct H. 
  eapply pshare_join_full_false1.
  exists x. apply join_com; eauto.
Qed.

Lemma pshare_join_full_false3: forall (p1: pshare) sh3,
  join (proj1_sig p1) (proj1_sig pfullshare) sh3 -> False.
Proof.
  intros.
  eapply pshare_join_full_false1.
  eauto.
Qed.

Lemma pshare_join_full_false4: forall (p1: pshare) sh3,
  join (proj1_sig pfullshare) (proj1_sig p1) sh3 -> False.
Proof.
  intros.
  eapply pshare_join_full_false1.
  eexists; eauto.
Qed.


Lemma pshare_pjoin_full_false3: forall (p1: pshare) sh3,
  pjoin p1 pfullshare sh3 -> False.
Proof.
  intros.
  apply (pshare_join_full_false3 p1 (pshare_sh sh3)).
  destruct p1; destruct sh3; unfold pjoin,pfullshare in H; simpl in *.
  auto.
Qed.

Lemma pshare_pjoin_full_false4: forall (p1: pshare) sh3,
  pjoin pfullshare p1 sh3 -> False.
Proof.
  intros.
  apply (pshare_join_full_false4 p1 (pshare_sh sh3)).
  destruct p1; destruct sh3; unfold pjoin,pfullshare in H; simpl in *.
  auto.
Qed.


Ltac pfullshare_join := 
  elimtype False; 
  solve [ eapply pshare_join_full_false1; eauto
    | eapply pshare_join_full_false2; eauto
    | eapply pshare_join_full_false3; eauto
    | eapply pshare_join_full_false4; eauto
    | eapply pshare_pjoin_full_false3; eauto
    | eapply pshare_pjoin_full_false4; eauto
  ].

Lemma rel_congruence : forall a x1 x2,
  join_sub x1 x2 ->
  join_sub (rel a x1) (rel a x2).
Proof.
  intros.
  destruct H.
  exists (rel a x).
  apply rel_join; auto.
Qed.

Lemma join_sub_pfullshare:
  forall (p: pshare), join_sub (proj1_sig pfullshare) (proj1_sig p) -> p=pfullshare.
Proof.
  intros.
  destruct H as [sh ?].
  unfold pfullshare in *.
  destruct p.
  assert (x = top).
  destruct (top_correct' x).
  destruct (join_assoc H H0) as [f [? ?]].
  generalize (split_identity _ _ H1 (unit_identity _ _ (join_com H2))); intro.
  assert (joins sh top) by eauto.
  generalize (identity_unit H3 H4); intro.
  unfold unit_for in H5.
  eapply join_eq.
  eauto.
  apply join_com; auto.
  subst.
  rewrite (proof_irr n top_share_nonidentity).
  auto.
Qed.

Lemma share_split_nonidentity1: forall sh: pshare,  nonidentity (fst (Share.split (pshare_sh sh))).
Proof.
  intros.
  case_eq (Share.split (pshare_sh sh)).
  intros.
  generalize (split_nontrivial' _ _ _ H); intro.
  intro.
  simpl in *.
  spec H0; auto.
  assert (nonidentity (pshare_sh sh)).
  destruct sh; simpl; auto.
  contradiction.
Qed.

Lemma share_split_nonidentity2: forall sh: pshare,  nonidentity (snd (Share.split (pshare_sh sh))).
Proof.
  intros.
  case_eq (Share.split (pshare_sh sh)).
  intros.
  generalize (split_nontrivial' _ _ _ H); intro.
  intro.
  simpl in *.
  spec H0; auto.
  assert (nonidentity (pshare_sh sh)).
  destruct sh; simpl; auto.
  contradiction.
Qed.

Definition split_pshare (sh: pshare) : pshare * pshare :=
  (mk_pshare _ (share_split_nonidentity1 sh), mk_pshare _ (share_split_nonidentity2 sh)).

Lemma mk_pshare_refl1: forall sh p1 p2, mk_pshare sh p1 = mk_pshare sh p2.
Proof.
  intros; rewrite (proof_irr p1 p2); auto.
Qed.

Lemma share_split_injective:
  forall sh1 sh2, Share.split sh1 = Share.split sh2 -> sh1=sh2.
Proof.
  intros sh1 sh2;
    case_eq (Share.split sh1); case_eq (Share.split sh2); intros.
  generalize (split_join _ _ _ H); intro.
  generalize (split_join _ _ _ H0); intro.
  inv H1.
  eapply join_eq; eauto.
Qed.

Lemma pshare_ext: forall sh1 sh2, pshare_sh sh1 = pshare_sh sh2 -> sh1=sh2.
Proof.
  destruct sh1; destruct sh2; simpl; intros; subst.
  rewrite (proof_irr n n0); auto.
Qed.

Lemma pshare_eq_dec: forall sh1 sh2: pshare, {sh1=sh2}+{sh1<>sh2}.
Proof.
  intros.
  destruct sh1; destruct sh2.
  destruct (Share.eq_dec x x0); [left|right].
  subst x0.
  rewrite (proof_irr n n0); auto.
  intro.
  inv H.
  contradiction n1; auto.
Qed.

Lemma share_joins_constructive:
  forall sh1 sh2 : t , joins sh1 sh2 ->  {sh3 | join sh1 sh2 sh3}.
Proof.
  intros.
  exists (lub sh1 sh2).
  destruct H.
  destruct H; split; auto.
Qed.

Lemma share_join_sub_constructive:
  forall sh1 sh3 : t , join_sub sh1 sh3 ->  {sh2 | join sh1 sh2 sh3}.
Proof.
  intros.
  exists (glb sh3 (comp sh1)).
  destruct H.
  destruct H.
  split.
  rewrite (glb_commute sh3 (comp sh1)).
  rewrite <- glb_assoc.
  rewrite comp2.
  rewrite glb_commute.
  rewrite glb_bot.
  auto.
  rewrite distrib2.
  rewrite comp1.
  rewrite glb_top.
  rewrite <- ord_spec2.
  rewrite <- H0.
  apply lub_upper1.
Qed.

Lemma triple_join_exists_share : @triple_join_exists t sa.
Proof.
  repeat intro.
  destruct H; destruct H0; destruct H1.
  exists (lub a (lub b c)).
  split.
  rewrite <- H2.
  rewrite glb_commute.
  rewrite distrib1.
  rewrite glb_commute.
  rewrite H1.
  rewrite glb_commute.
  rewrite H0.
  rewrite lub_bot; auto.
  rewrite <- H2.
  apply lub_assoc.
Qed.

Lemma pshare_split_neq1: forall (sh: pshare) sh1 sh2, split sh = (sh1, sh2) -> sh1 <> sh.
Proof.
intros until sh2; intros H contra.
assert (H2: nonidentity (snd (split sh))) by (apply share_split_nonidentity2).
rewrite H in H2.
simpl in *.
apply split_join in H.
rewrite contra in H.
apply join_com in H.
apply unit_identity in H.
apply H2; auto.
Qed.

Lemma pshare_split_neq2: forall (sh: pshare) sh1 sh2, split sh = (sh1, sh2) -> sh2 <> sh.
Proof.
intros until sh2; intros H contra.
assert (H1: nonidentity (fst (split sh))) by (apply share_split_nonidentity1).
rewrite H in H1.
simpl in *.
apply split_join in H.
rewrite contra in H.
apply unit_identity in H.
apply H1; auto.
Qed.

Lemma nonemp_split_neq1: forall sh sh1 sh2, nonidentity sh -> split sh = (sh1, sh2) -> sh1 <> sh.
Proof.
intros until sh2; intros H H0.
assert (H1: sh1 <> proj1_sig (mk_pshare sh H)).
apply pshare_split_neq1 with (sh2 := sh2).
simpl; auto.
simpl in H1; auto.
Qed.

Lemma nonemp_split_neq2: forall sh sh1 sh2, nonidentity sh -> split sh = (sh1, sh2) -> sh2 <> sh.
Proof.
intros until sh2; intros H H0.
assert (H1: sh2 <> proj1_sig (mk_pshare sh H)).
apply pshare_split_neq2 with (sh1 := sh1).
simpl; auto.
simpl in H1; auto.
Qed.




Module Type ShareMapInput.
  Variable A:Type.
  Variable A_dec_eq : forall x y:A, {x=y}+{x<>y}.

  Variable B:Type.
End ShareMapInput.


Module ShareMap (I:ShareMapInput).
  Import I.
  
  Definition map := fpm A (I.B * Share.t) (sa_prod (sa_equiv B) Share.sa).
  Definition map_sa := fpm_sa I.A I.A_dec_eq (I.B * Share.t) (sa_prod (sa_equiv B) Share.sa).
  Existing Instance map_sa.

  Definition map_share (a:A) (m:map) : Share.t :=
    match lookup_fpm _ _ _ m a with
    | Some (exist z H) => snd z
    | None => Share.bot
    end.

  Definition map_val (a:A) (m:map) : option B :=
    match lookup_fpm _ _ _ m a with
    | Some (exist z H) => Some (fst z)
    | None => None
    end.

  Definition empty_map : map := empty_fpm _ _ _.

  Program Definition map_upd (a:A) (b:B) (m:map) : option map :=
    match lookup_fpm _ _ _ m a with
    | Some (exist z H) =>
        if Share.eq_dec (snd z) Share.top
           then Some (insert_fpm _ A_dec_eq _ _ a (exist _ (b,snd z) _) m)
           else None
    | None => None
    end.
  Next Obligation.
    simpl in *; subst.
    intro.
    apply Share.nontrivial.
    apply identity_share_bot.
    rewrite identity_unit_equiv.
    rewrite identity_unit_equiv in H0.
    destruct H0; auto.
  Qed.

  Lemma map_join_char : forall m1 m2 m3,
    join m1 m2 m3 <->
    (forall a,
       join (map_share a m1)
            (map_share a m2)
            (map_share a m3) /\

          ((map_val a m1 = map_val a m2 /\
            map_val a m2 = map_val a m3)\/
          (map_val a m1 = None /\
           map_val a m2 = map_val a m3) \/
          (map_val a m1 = map_val a m3 /\
           map_val a m2 = None))).
  Proof.
    split; intros.
    hnf in H. spec H a.
    inv H. split.
    unfold map_share.
    unfold lookup_fpm.
    rewrite <- H1.
    rewrite H3.
    destruct (proj1_sig m3 a). destruct p.
    split.
    rewrite Share.glb_commute.
    apply Share.glb_bot.
    rewrite Share.lub_commute.
    apply Share.lub_bot.
    split.
    apply Share.glb_bot.
    apply Share.lub_bot.
    right. left.
    unfold map_val, lookup_fpm.
    rewrite <- H1.
    rewrite H3.
    split; auto.
    split.
    unfold map_share, lookup_fpm.
    rewrite <- H2.
    rewrite H3.
    destruct (proj1_sig m3 a).
    destruct p.
    split. apply Share.glb_bot. apply Share.lub_bot.
    split. apply Share.glb_bot. apply Share.lub_bot.
    right. right.
    unfold map_val. unfold lookup_fpm.
    rewrite H3. rewrite <- H2.
    split; auto.
    split.
    unfold map_share, lookup_fpm.
    rewrite <- H0. rewrite <- H1. rewrite <- H2.
    destruct x; destruct y; destruct z.
    destruct H3; auto.
    left.
    unfold map_val, lookup_fpm.
    rewrite <- H0. rewrite <- H1. rewrite <- H2.
    destruct x; destruct y; destruct z.
    destruct H3; auto.
    inv H. intuition congruence.

    intro x. spec H x.
    destruct H.
    destruct H0.
    destruct H0.
    unfold map_val in *.
    unfold map_share in *.
    unfold lookup_fpm in *.
    destruct (proj1_sig m1 x).
    destruct (proj1_sig m2 x).
    destruct (proj1_sig m3 x).
    destruct p; destruct p0; destruct p1.
    constructor.
    simpl. split; auto.
    inv H0; inv H1; inv H2.
    split; auto.
    destruct p; destruct p0. discriminate.
    destruct p; discriminate.
    destruct (proj1_sig m2 x).
    destruct (proj1_sig m3 x).
    destruct p; discriminate.
    destruct p; discriminate.
    destruct (proj1_sig m3 x).
    destruct p; discriminate.
    constructor.
    
    destruct H0.
    destruct H0.
    unfold map_val, map_share, lookup_fpm in *.
    destruct (proj1_sig m1 x); try discriminate.
    destruct p. discriminate.
    destruct (proj1_sig m2 x);
      destruct (proj1_sig m3 x).
    destruct p; destruct p0. 
    inv H1. 
    assert (x0 = x1).
    destruct x0; destruct x1. simpl in *.
    f_equal; auto.
    destruct H.
    rewrite Share.lub_commute in H1.
    rewrite Share.lub_bot in H1.
    auto.
    subst x1.
    replace n0 with n. constructor.
    apply proof_irr.
    destruct p; discriminate.
    destruct p; discriminate.
    constructor.

    unfold map_share, map_val, lookup_fpm in *.
    destruct H0.
    destruct (proj1_sig m2 x).
    destruct p; try discriminate.
    destruct (proj1_sig m1 x);
      destruct (proj1_sig m3 x).
    destruct p; destruct p0.
    assert (x0 = x1).
    destruct x0; destruct x1.
    simpl in *.
    f_equal. injection H0; auto.
    destruct H.
    rewrite Share.lub_bot in H2. auto.
    subst x1.
    replace n0 with n. constructor.
    apply proof_irr.
    destruct p; discriminate.
    destruct p; discriminate.
    constructor.
  Qed.

  Lemma empty_map_identity : identity empty_map.
  Proof.
    rewrite identity_unit_equiv.
    intro x. simpl. constructor.
  Qed.
    
  Lemma map_identity_unique : forall m1 m2:map,
    identity m1 -> identity m2 -> m1 = m2.
  Proof.
    intros.
    destruct m1; destruct m2; simpl in *.
    cut (x = x0). intros. subst x0.
    replace f0 with f; auto.
    apply proof_irr; auto.
    rewrite identity_unit_equiv in H, H0.
    extensionality a.
    spec H a; spec H0 a. simpl in *.
    inv H; inv H0; auto; try discriminate;
      elimtype False; apply (pnjoin_self _ _ x1 x1); auto.
  Qed.

  Lemma map_identity_is_empty : forall m,
    identity m -> m = empty_map.
  Proof.
    intros; apply map_identity_unique; auto.
    apply empty_map_identity.
  Qed.

  Lemma empty_map_join : forall m,
    join empty_map m m.
  Proof.
    intro m. destruct (join_ex_units m).
    replace empty_map with x; auto.
    apply map_identity_is_empty.
    eapply unit_identity; eauto.
  Qed.

  Lemma map_val_bot : forall a m,
    map_val a m = None <-> map_share a m = Share.bot.
  Proof.
    do 2 intro.
    unfold map_val, map_share, lookup_fpm.
    destruct (proj1_sig m a); intuition.
    destruct p. discriminate.
    destruct p. elim n.
    rewrite identity_unit_equiv.
    split.
    hnf; auto.
    rewrite H.
    split.
    apply Share.glb_bot. apply Share.lub_bot.
  Qed.

  Lemma map_upd_success : forall a v m,
    map_share a m = Share.top ->
    exists m', map_upd a v m = Some m'.
  Proof.
    intros.
    unfold map_upd.
    generalize (refl_equal (lookup_fpm _ _ _ m a)).
    pattern (lookup_fpm _ _ _ m a) at 1 3.
    case (lookup_fpm _ _ _ m a).
    do 2 intro.
    destruct p; destruct x.
    simpl.
    destruct (Share.eq_dec t0 Share.top); intros; try discriminate.
    econstructor. reflexivity.
    unfold map_share in H.
    rewrite <- e in H.
    simpl in H. injection H; contradiction.
    intros.
    unfold map_share in H.
    rewrite <- e in H.
    elim Share.nontrivial. auto.
  Qed.

  Lemma map_set_share1 : forall a v m m',
    map_upd a v m = Some m' ->
    map_share a m = Share.top.
  Proof.
    do 4 intro.
    unfold map_upd.
    generalize (refl_equal (lookup_fpm _ _ _ m a)).
    pattern (lookup_fpm _ _ _ m a) at 1 3.
    case (lookup_fpm _ _ _ m a).
    do 2 intro.
    destruct p; destruct x.
    simpl.
    destruct (Share.eq_dec t0 Share.top); intros; try discriminate.
    injection H; clear H; intro; subst m'.
    unfold map_share.
    rewrite <- e.
    simpl; f_equal. auto.
    intros; discriminate.
  Qed.

  Lemma map_set_share2 : forall a v m m',
    map_upd a v m = Some m' ->
    map_share a m' = Share.top.
  Proof.
    do 4 intro.
    unfold map_upd.
    generalize (refl_equal (lookup_fpm _ _ _ m a)).
    pattern (lookup_fpm _ _ _ m a) at 1 3.
    case (lookup_fpm _ _ _ m a).
    do 2 intro.
    destruct p; destruct x.
    simpl.
    destruct (Share.eq_dec t0 Share.top); intros; try discriminate.
    injection H; clear H; intro; subst m'.
    unfold map_share.
    rewrite fpm_gss.
    simpl; f_equal. auto.
    intros; discriminate.
  Qed.

  Lemma map_set_share3 : forall a v m m',
    map_upd a v m = Some m' ->
    forall a',
      map_share a' m = map_share a' m'.
  Proof.
    intros.
    destruct (A_dec_eq a a'). subst a'.
    erewrite map_set_share1; eauto.
    erewrite map_set_share2; eauto.
    revert H.
    unfold map_upd.
    generalize (refl_equal (lookup_fpm _ _ _ m a)).
    pattern (lookup_fpm _ _ _ m a) at 1 3.
    case (lookup_fpm _ _ _ m a).
    do 2 intro.
    destruct p; destruct x.
    simpl.
    destruct (Share.eq_dec t0 Share.top); intros; try discriminate.
    injection H; clear H; intro; subst m'.
    unfold map_share at 2.
    rewrite fpm_gso; auto.
    intros; discriminate.
  Qed.

  Lemma map_gss_val: forall a v m m',
        map_upd a v m = Some m' ->
        map_val a m' = Some v.
  Proof.
    do 4 intro.
    unfold map_upd.
    generalize (refl_equal (lookup_fpm _ _ _ m a)).
    pattern (lookup_fpm _ _ _ m a) at 1 3.
    case (lookup_fpm _ _ _ m a).
    do 2 intro.
    destruct p; destruct x.
    simpl.
    destruct (Share.eq_dec t0 Share.top); intros.
    inversion H.
    unfold map_val.
    rewrite fpm_gss.
    simpl. auto.
    discriminate.
    intros. discriminate.
  Qed.

  Lemma map_gso_val : forall i j v m m', 
       i <> j ->
       map_upd j v m = Some m' ->
       map_val i m = map_val i m'.
  Proof.
    do 6 intro.
    unfold map_upd.
    generalize (refl_equal (lookup_fpm _ _ _ m j)).
    pattern (lookup_fpm _ _ _ m j) at 1 3.
    case (lookup_fpm _ _ _ m j).
    do 2 intro.
    destruct p; destruct x.
    simpl.
    destruct (Share.eq_dec t0 Share.top); intros; try discriminate.
    injection H0. intro.
    subst m'. clear  H0.
    generalize (map_upd_obligation_1 j v m (b,t0) n e e0).
    subst t0; intros.
    unfold map_val.
    rewrite fpm_gso; auto.
    intros; discriminate.
  Qed.

  Lemma map_gso_share : forall i j v m m',
    i <> j ->
    map_upd j v m = Some m' ->
    map_share i m = map_share i m'.
  Proof.
    do 6 intro.
    unfold map_upd.
    generalize (refl_equal (lookup_fpm _ _ _ m j)).
    pattern (lookup_fpm _ _ _ m j) at 1 3.
    destruct (lookup_fpm _ _ _ m j). intros.
    destruct p.
    destruct (eq_dec (snd x) top).
    inv H0.
    unfold map_share.
    rewrite fpm_gso; auto.
    discriminate.
    intros. discriminate.
  Qed.    

  Lemma map_upd_join : forall m1 m2 m3 a v m1',
    map_upd a v m1 = Some m1' ->
    join m1 m2 m3 ->
    exists m3', map_upd a v m3 = Some m3' /\
      join m1' m2 m3'.
  Proof.
    intros.
    rewrite map_join_char in H0.
    destruct (H0 a).
    generalize H; intros.
    apply map_set_share1 in H.
    rewrite H in H1.
    destruct H1.
    rewrite glb_commute in H1.
    rewrite glb_top in H1.
    rewrite H1 in H4.
    rewrite lub_bot in H4.
    symmetry in H4.
    destruct (map_upd_success a v _ H4).
    exists x; split; auto.
    clear H2.
    rewrite map_join_char.
    intro a'.
    destruct (A_dec_eq a a').
    subst a'. split.
    apply map_set_share2 in H3. rewrite H3.
    apply map_set_share2 in H5. rewrite H5.
    rewrite H1.
    split. apply glb_bot. apply lub_bot.
    right. right.
    split.
    erewrite map_gss_val; eauto.
    erewrite map_gss_val; eauto.
    rewrite map_val_bot. auto.
    destruct (H0 a'). split.
    rewrite <- (map_gso_share a' a v m1 m1'); auto.
    rewrite <- (map_gso_share a' a v m3 x); auto.
    rewrite <- (map_gso_val a' a v m1 m1'); auto.
    rewrite <- (map_gso_val a' a v m3 x); auto.
  Qed.

  Program Definition build_map (l:list (A * B)) :=
     fold_right (fun (ab:A * B) m =>
                     insert_fpm A A_dec_eq (B * Share.t) (sa_prod (sa_equiv B) Share.sa) (fst ab)
                     (exist (fun x => nonidentity x) (snd ab, Share.top) _) m)
      (empty_fpm A (B * Share.t)%type _) l.
  Next Obligation.
    clear l.
    simpl. intro.
    apply Share.nontrivial.
    apply identity_share_bot; auto.
    rewrite identity_unit_equiv in H.
    destruct H. simpl in *.
    rewrite identity_unit_equiv; auto.
  Defined.

  Lemma build_map_results : forall (l:list (A*B)) a b,
    NoDup (List.map (@fst _ _) l) ->
    (In (a,b) l <->
    (map_val a (build_map l) = Some b /\
     map_share a (build_map l) = Share.top)).
  Proof.
    induction l; simpl.
    split; intros. elim H0.
    destruct H0.
    unfold build_map in H0. simpl in H0.
    unfold map_val in H0.
    simpl in H0. discriminate.
    intros. split; intros.
    destruct H0; subst.
    unfold build_map.
    simpl fold_right.
    split.
    unfold map_val.
    rewrite fpm_gss. simpl; auto.
    unfold map_share.
    rewrite fpm_gss. simpl; auto.
    generalize H0; intro H1.
    rewrite IHl in H0.
    inv H.
    assert (fst a <> a0).
    intro. subst a0.
    elim H4.
    clear -H1. induction l; simpl in *; intuition; subst; auto.
    destruct H0.
    unfold build_map. simpl fold_right.
    split.
    unfold map_val.
    rewrite fpm_gso; auto.
    unfold map_share.
    rewrite fpm_gso; auto.
    inv H. auto.
    inv H.
    destruct H0.
    destruct a.
    destruct (A_dec_eq a a0).
    subst a0.
    left. f_equal.
    unfold build_map in H.
    unfold map_val in H.
    simpl fold_right in H.
    rewrite fpm_gss in H.
    inv H. auto.
    right.
    rewrite IHl; auto.
    split.
    revert H.
    unfold build_map, map_val.
    simpl fold_right.
    rewrite fpm_gso; auto.
    revert H0.
    unfold build_map, map_share.
    simpl fold_right.
    rewrite fpm_gso; auto.
  Qed.

  Lemma build_map_join : forall (l1 l2:list (A * B)),
    NoDup (List.map (@fst _ _) (l1++l2)) ->
    join  (build_map l1)
          (build_map l2)
          (build_map (l1++l2)).
  Proof.
    induction l1; intros.
    simpl app. 
    unfold build_map at 1.
    simpl fold_right.
    apply empty_fpm_join.
    inv H.
    simpl app.
    unfold build_map.
    simpl fold_right.
    apply insert_fpm_join.
    2: apply (IHl1 l2); auto.
    assert (~In (fst a) (List.map (@fst _ _) l2)).
    intro.
    elim H2.
    rewrite map_app.
    apply in_or_app.
    auto.
    clear -H.
    induction l2; simpl in *.
    auto.
    rewrite fpm_gso; auto.
  Qed.

End ShareMap.
