(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import ageable.
Require Import predicates_hered.

Local Open Scope pred.

Lemma sub_equ `{ageable A} : forall G P Q,
  G |-- P >=> Q ->
  G |-- Q >=> P ->
  G |-- P <=> Q.
Proof.
  repeat intro.
  split.
  eapply H0; eauto.
  eapply H1; eauto.
Qed.

Lemma equ_sub `{ageable A} : forall G P Q,
  G |-- P <=> Q ->
  G |-- P >=> Q.
Proof.
  repeat intro.
  destruct (H0 a H1 a' H2); eauto.
Qed.

Lemma equ_sub2 `{ageable A} : forall G P Q,
  G |-- P <=> Q ->
  G |-- Q >=> P.
Proof.
  repeat intro.
  destruct (H0 a H1 a' H2); eauto.
Qed.

Lemma equ_comm : forall `{ageable A} (P Q:pred A),
  P <=> Q = Q <=> P.
Proof.
  intros. apply equiv_eq.
  apply sub_equ.
    apply equ_sub2. hnf; auto.
    apply equ_sub. hnf; auto.
  apply sub_equ.
    apply equ_sub2. hnf; auto.
    apply equ_sub. hnf; auto.
Qed.    

Lemma sub_refl `{ageable A} : forall G P,
  G |-- P >=> P.
Proof.
  repeat intro; auto.
Qed.

Lemma sub_trans `{ageable A} : forall G P Q R,
  G |-- P >=> Q ->
  G |-- Q >=> R ->
  G |-- P >=> R.
Proof.
  repeat intro.
  eapply H1; eauto.
  eapply H0; eauto.
Qed.

Lemma sub_top `{ageable A} : forall G P,
  G |-- P >=> TT.
Proof.
  repeat intro; simpl; auto.
Qed.

Lemma sub_bot `{ageable A} : forall G P,
  G |-- FF >=> P.
Proof.
  repeat intro; simpl in *; intuition.
Qed.

Lemma sub_conj `{ageable A} : forall G P P' Q Q',
  G |-- P >=> P' ->
  G |-- Q >=> Q' ->
  G |-- P && Q >=> (P' && Q').
Proof.
  repeat intro.
  destruct H5; split.
  eapply H0; eauto.
  eapply H1; eauto.
Qed.

Lemma sub_impl `{ageable A} : forall G P P' Q Q',
  G |-- P' >=> P ->
  G |-- Q >=> Q' ->
  G |-- (P --> Q) >=> (P' --> Q').
Proof.
  repeat intro.
  eapply H1; eauto.
  apply H5; auto.
  eapply H0; eauto.
Qed.

Lemma sub_disj `{ageable A} : forall G P P' Q Q',
  G |-- P >=> P' ->
  G |-- Q >=> Q' ->
  G |-- (P || Q) >=> (P' || Q').
Proof.
  repeat intro.
  destruct H5; [ left | right ].
  eapply H0; eauto.
  eapply H1; eauto.
Qed.

Lemma sub_forall `{ageable A} : forall G B (X Y:B -> pred A),
  (forall x:B, G |-- X x >=> Y x) ->
  G |-- universal X >=> universal Y.
Proof.
  repeat intro.
  eapply H0; eauto.
Qed.

Lemma sub_exists `{ageable A} : forall G B (X Y:B -> pred A),
  (forall x:B, G |-- X x >=> Y x) ->
  G |-- existential X >=> existential Y.
Proof.
  repeat intro.
  destruct H4; exists x.
  eapply H0; eauto.
Qed.

Lemma sub_forall_spec `{ageable A} : forall G B (X:B -> pred A) x,
  G |-- universal X >=> X x.
Proof.
  repeat intro; eauto.
Qed.

Lemma sub_exists_spec `{ageable A} : forall G B(X:B -> pred A) x,
  G |-- X x >=> existential X.
Proof.
  repeat intro.
  exists x; auto.
Qed.

Lemma sub_later `{ageable A} : forall G P Q,
  G |-- |>(P >=> Q) <-> G |-- |>P >=> |>Q.
Proof.
  intros.
  rewrite <- later_commute.
  rewrite later_impl.
  tauto.
Qed.

Lemma sub_box `{ageable A} : forall G (M:modality) P Q,
  inclusion A M fashionR ->
  G |-- P >=> Q ->
  G |-- box M P >=> box M Q.
Proof.
  repeat intro.
  generalize (H0 _ _ H6); intros.
  rewrite <- (box_refl_trans fashionM) in H1.
  apply H1 in H2.
  spec H2 a' H3.
  eapply pred_nec_hereditary in H2; eauto.
  spec H2 a'1 H7.
  apply H2; auto.
  simpl; apply fashionR_refl.
  simpl; exact fashionR_trans.
Qed.

Lemma sub_diamond `{ageable A} : forall G (M:modality) P Q,
  inclusion A M fashionR ->
  G |-- P >=> Q ->
  G |-- diamond M P >=> diamond M Q.
Proof.
  repeat intro.
  destruct H5 as [w [? ?]].
  exists w; split; auto.
  rewrite <- (box_refl_trans fashionM) in H1.
  apply H1 in H2.
  spec H2 a' H3.
  eapply pred_nec_hereditary in H2; eauto.
  assert (fashionR a'0 w).
  apply fashionR_sym.
  apply H0; auto.
  eapply H2; eauto.
  simpl; apply fashionR_refl.
  simpl; exact fashionR_trans.
Qed.
